#ifndef DICT_H
#define DICT_H

#include <QPair>
#include <QList>
#include <QStringList>
#include <QTimer>
#include <QByteArray>
#include <QTcpSocket>

typedef QPair<QString, QString> DictPair;
typedef QList<DictPair> DictList;

struct DictRequest {
	enum RequestCommand {
		None,
		Define,
		Match,
		Databases,
		Strategies,
		Info,
		Server,
		Option,
		Client,
		Auth,
		Status,
		Help,
		Quit
	};
	DictRequest(RequestCommand req, const QString &cmd)
		: type(req), command(cmd) {}
	RequestCommand type;
	QString command;
};

class DictResponse : public QObject {
	Q_OBJECT

public:
	enum ResponseCode {
		DatabaseList = 110,
		StrategyList = 111,
		DatabaseInfo = 112,
		HelpMessage = 113,
		ServerInfo = 114,
		Definitions = 150,
		Words = 151,
		Matches = 152,

		StatusMessage = 210,
		Hello = 220,
		Bye = 221,
		AuthOK = 230,
		OK = 250,

		ServerUnavailable = 420,
		ServerShutdown = 421,

		UnrecognizedCommand = 500,
		IllegalParameter = 501,
		UnimplementedCommand = 502,
		UnimplementedParameter = 503,
		AccessDenied = 530,
		AuthDenied = 531,
		UnknownMechanism = 532,
		InvalidDatabase = 550,
		InvalidStrategy = 551,
		NoMatch = 552,
		NoDatabase = 554,
		NoStrategy = 555,
	};
	ResponseCode type;
	QString message, text;

	DictResponse(ResponseCode code) : type(code) {}

	virtual void parse(const QString &line) {}
	virtual void append(const QString &line) { text += line + '\n'; }
	virtual void end() {}
	virtual bool done() const { return true; }

Q_SIGNALS:
	void processed(const DictResponse &);
};

class ListResponse : public DictResponse {
public:
	ListResponse(ResponseCode code) : DictResponse(code) {}
	DictList list;

	void append(const QString &line);
};

class MatchResponse : public DictResponse {
public:
	MatchResponse(ResponseCode code) : DictResponse(code), isDone(false) {}
	QString database;
	QStringList list;
	bool isDone;

	void append(const QString &line);
	bool done() const { return isDone; }
};

class DefineResponse : public MatchResponse {
public:
	DefineResponse(ResponseCode code) : MatchResponse(code) {}
	QString description;

	void parse(const QString &line);
	void append(const QString &line) { text += line + '\n'; }
	void end() { list << text; }
};

class MimeResponse : public DefineResponse {
public:
	MimeResponse(ResponseCode code) : DefineResponse(code) {}
	QStringList header;
	bool isHeader;

	void parse(const QString &line);
	void append(const QString &line);
	void end();
	void decode();
};

typedef QList<const DictRequest *> DictQueue;

class Dict : public QObject {
	Q_OBJECT

public:
	static const char *DefaultHost;
	static const int DefaultPort;
	static const char *DefaultDatabase;
	static const char *DefaultStrategy;
	static const char *AllDatabases;

	Dict(const QString &host = DefaultHost, int port = DefaultPort);
	~Dict();
	void setHost(const QString &hostname, int port = DefaultPort);
	void setUser(const QString &username, const QString &password);
	void setClient(const QString &client);
	enum DictOption { Mime = 1 };
	void setOption(DictOption opt, bool on);

	void setTimeouts(int connect = 20, int read = 10, int idle = 300);
	void setKeepConnect(bool on) { keepconnect = on; }
	void setKeepAlive(bool on) { keepalive = on; };

	QString host() const { return dict.host; }
	int port() const { return dict.port; }
	QString client() const { return dict.client; }
	QString user() const { return dict.user; }
	QString ident() const { return dict.ident; }
	QStringList capabilities() const { return dict.capabilities; }

	enum NetworkError {
		UnknownError,
		HostNotFound = 600,
		ConnectionRefused,
		ConnectTimeout,
		ReadTimeout,
	};

public Q_SLOTS:
	void define(const QString &query, const QString &database = DefaultDatabase);
	void match(const QString &query, const QString &database = DefaultDatabase, const QString &strategy = DefaultStrategy);
	void info(const QString &query);
	void auth(const QString &username, const QString &password);
	void status();
	void quit();
	void cancel();

private Q_SLOTS:
	void read();
	void socketError(QAbstractSocket::SocketError);
	void socketState(QAbstractSocket::SocketState);
	void connectTimeout();
	void readTimeout();

Q_SIGNALS:
	void responseReceived(const DictResponse &);
	void responseProcessed(const DictResponse &);
	void responseText(const QString &);
	void connected(bool);
	void authenticated(bool);
	void identified(bool);
	void statusMessage(const QString &);
	void errorMessage(const QString &, int code = 0);

private:
	enum DictState {
		Disconnected,
		HostLookup,
		Connecting,
		Connected,
		Writing,
		Reading,
		Parsing,
		Appending,
		Canceling,
	};
	QTcpSocket *socket;

	struct DictInfo {
		QString host;
		int port;
		QString client;
		QString user, password;
		QString ident, msgid;
		QStringList capabilities;
		bool mime;
		void clear() { ident.clear(); msgid.clear(); capabilities.clear(); }
	} dict;

	QTimer connectTimer, readTimer, idleTimer;

	DictState state;
	DictResponse *response, *nomatch;
	DictQueue sendqueue, recvqueue;
	QByteArray extra;
	bool interauth;
	bool keepconnect, keepalive;

	void setupSocket();
	void setupTimers();
	void reset();
	void abort();
	void clear();

	void open();
	void auth();
	void client();
	void option(DictOption opt);
	void serverIdent(const QString &text);

	void send(const DictRequest *request);
	void read(const QString &text);
	void parse(const QString &text);
	void done();

	DictResponse *newResponse(DictResponse::ResponseCode code);
	void delResponse(void);
	void okResponse(void);
	void helloResponse(void);
};

#endif
