#include "dictbrowser.h"
#include "dictpage.h"

#include <QScrollBar>


DictBrowser::DictBrowser(QWidget *parent)
	: QTextBrowser(parent), histSize(100), plaintext(false), hiliter(0)
{
	setOpenLinks(false);
	page = new DictPage();
	entry.type = 0;
	connect(this, SIGNAL(highlighted(const QString &)),
		SLOT(pointCursor(const QString &)));
	connect(this, SIGNAL(cursorPositionChanged()), SLOT(selection()));
	connect(page, SIGNAL(changed(bool)), SLOT(setPage(bool)));
}

DictBrowser::~DictBrowser()
{
	delete page;
}

void DictBrowser::setPage(bool append)
{
	if (!append)
		setPage(page->title(), page->htmltext());
	else
		appendPage(page->title(), page->htmltext());
}

void DictBrowser::moveTo(int steps)
{
	QList<HistoryEntry> &history = (steps > 0)
		? forwardHistory : backwardHistory;

	if (history.isEmpty())
		return;

	QList<HistoryEntry> &other = (steps > 0)
		? backwardHistory : forwardHistory;

	savePosition();
	if (steps < 0)
		steps = -steps;
	while (steps--) {
		other.prepend(entry);
		entry = history.takeFirst();
	}
	displayPage(entry.page);
	restorePosition();

	emit backwardAvailable(!backwardHistory.isEmpty());
	emit forwardAvailable(!forwardHistory.isEmpty());
	emit historyChanged(entry.title);
}

void DictBrowser::savePosition()
{
	entry.hpos = horizontalScrollBar()->value();
	entry.vpos = verticalScrollBar()->value();
}

void DictBrowser::restorePosition()
{
	horizontalScrollBar()->setValue(entry.hpos);
	verticalScrollBar()->setValue(entry.vpos);
}

void DictBrowser::clearHistory()
{
	backwardHistory.clear();
	forwardHistory.clear();
	emit backwardAvailable(false);
	emit forwardAvailable(false);
}

QStringList DictBrowser::historyList(const QList<HistoryEntry> &history, int len) const
{
	QStringList list;
	len = qMin(history.size(), len);
	for (int i = 0; i < len; ++i)
		list << history[i].title;
	return list;
}

void DictBrowser::setPage(const QString &query, const QString &text)
{
	savePosition();
	displayPage(text);

	if ((entry.title != query || entry.type != page->type())
	&& !entry.title.isEmpty()) {
		backwardHistory.prepend(entry);
		forwardHistory.clear();
	}
	if (backwardHistory.size() > histSize)
		backwardHistory.removeLast();

	entry.title = query;
	entry.page = text;
	entry.type = page->type();

	emit backwardAvailable(!backwardHistory.isEmpty());
	emit forwardAvailable(!forwardHistory.isEmpty());
	emit historyChanged(query);
}

void DictBrowser::appendPage(const QString &query, const QString &text)
{
	savePosition();
	if (plaintext) {
		moveCursor(QTextCursor::End);
		insertPlainText(text);
	} else append(text);
	restorePosition();
	entry.page += text;
}

void DictBrowser::displayPage(const QString &text)
{
	if (plaintext) setPlainText(text);
	else setHtml(text);
}

void DictBrowser::plainText(bool toggle)
{
	plaintext = toggle;
	displayPage(entry.page);
}

void DictBrowser::reload()
{
	savePosition();
	displayPage(entry.page);
	restorePosition();
}

QString resource(const QString &resfile);

QString DictBrowser::pageText() const
{
	QString html = resource("page.html")
		.arg(entry.title)
		.arg(document()->defaultStyleSheet())
		.arg(entry.page);
	return html;
}

void DictBrowser::setStyleSheet(const QString &style)
{
	document()->setDefaultStyleSheet(style);
	reload();
}

void DictBrowser::setTextFont(const QFont &font)
{
	document()->setDefaultFont(font);
	reload();
}

void DictBrowser::clearSelection()
{
	QTextCursor cursor = textCursor();
	cursor.clearSelection();
	setTextCursor(cursor);
}

void DictBrowser::busyCursor(bool busy)
{
	viewport()->setCursor((busy) ? Qt::BusyCursor : Qt::ArrowCursor);
	cursor = viewport()->cursor();
}

void DictBrowser::pointCursor(const QString &text)
{
	if (text.isEmpty())
		viewport()->setCursor(cursor);
}

void DictBrowser::focusOutEvent(QFocusEvent *event)
{
	QTextEdit::focusOutEvent(event);
}

#include <QContextMenuEvent>
#include <QMenu>

void DictBrowser::contextMenuEvent(QContextMenuEvent *event)
{
	QMenu *menu;

	if (actions().isEmpty())
		menu = createStandardContextMenu();
	else {
		menu = new QMenu();
		QAction *menubar = actions().first();
		if (!menubar->isChecked()) {
			menu->addAction(menubar);
			menu->addSeparator();
		}
		foreach (QAction *action, actions().mid(1))
			menu->addAction(action);
		if (actions().size() > 3)
			menu->insertSeparator(actions().at(3));
	}

	menu->exec(event->globalPos());
	delete menu;
}

#include <QPalette>

void DictBrowser::selection()
{
	if (!textCursor().hasSelection())
		return;

	QPalette palette;
	QColor text = palette.color(QPalette::HighlightedText);
	QColor back = palette.color(QPalette::Highlight);

	if (textBackgroundColor().value()
		&& textBackgroundColor().value() < textColor().value()) {
		palette.setColor(QPalette::HighlightedText, back);
		palette.setColor(QPalette::Highlight, text);
	} else {
		palette.setColor(QPalette::HighlightedText, text);
		palette.setColor(QPalette::Highlight, back);
	}
	setPalette(palette);
}


#include <KGlobalSettings>
#include <KLocale>

#include <QFileDialog>
#include <QPrinter>
#include <QPrintDialog>
#include <QPrintPreviewDialog>

void DictBrowser::savePage()
{
	static QString docPath = KGlobalSettings::documentPath();
	QString fileName = QFileDialog::getSaveFileName(this, i18n("Save Document"), docPath + "/" + entry.title + ".html");
	if (fileName.isEmpty())
		return;
	docPath = fileName.left(fileName.lastIndexOf('/'));

	QFile file(fileName);
	if (file.open(QFile::WriteOnly | QFile::Text)) {
		QTextStream out(&file);
		out.setCodec("UTF-8");
		out << pageText();
		emit highlighted(i18n("Document saved"));
	} else {
		page->setTitle(entry.title);
		page->error(i18n("Cannot write file <b>%1</b>: %2").arg(fileName, file.errorString()), DictPage::FileSystem);
	}
}

void DictBrowser::printPage()
{
	QPrinter printer;
	if (textCursor().hasSelection())
		printer.setPrintRange(QPrinter::Selection);
	QPrintDialog dialog(&printer, this);
	dialog.setWindowTitle(i18n("Print Document"));
	if (dialog.exec() == QDialog::Accepted)
		print(&printer);
}

void DictBrowser::previewPage()
{
	QPrinter printer;
	if (textCursor().hasSelection())
		printer.setPrintRange(QPrinter::Selection);
	QPrintPreviewDialog dialog(&printer, this);
	connect(&dialog, SIGNAL(paintRequested(QPrinter *)),
		SLOT(print(QPrinter *)));
	dialog.exec();
}

DictHiliter *DictBrowser::dictHiliter()
{
	if (!hiliter)
		hiliter = new DictHiliter(this);
	return hiliter;
}

void DictBrowser::setHilite(bool on, QTextCharFormat fmt)
{
	if (on) {
		dictHiliter();
		hiliter->setHiliteFormat(fmt, DictHiliter::Query);
		hiliter->setQuery(entry.title);
		connect(this, SIGNAL(historyChanged(const QString &)),
			hiliter, SLOT(setQuery(const QString &)));
	} else if (hiliter) {
		hiliter->setQuery("");
		disconnect(hiliter);
	}
}

#include <QTextCharFormat>

DictHiliter::DictHiliter(QTextEdit *parent)
	: QSyntaxHighlighter(parent)
{
	hilites.resize(2);

	QTextCharFormat format;
	format.setBackground(Qt::yellow);
	setHiliteFormat(format, Text);

	hilites[Text].regexp = QRegExp("", Qt::CaseInsensitive, QRegExp::FixedString);
	hilites[Query].regexp = QRegExp("", Qt::CaseInsensitive);
}

void DictHiliter::setText(const QString &text)
{
	hilites[Text].regexp.setPattern(text);
	rehighlight();
}

void DictHiliter::setQuery(const QString &text)
{
	hilites[Query].regexp.setPattern((text.size()) ? "\\b" + text : "");
	rehighlight();
}

void DictHiliter::setHiliteFormat(QTextCharFormat format, DictHilite hilite)
{
	hilites[hilite].format = format;
}

void DictHiliter::highlightBlock(const QString &text)
{
	foreach (const Hilite &hilite, hilites) {
		const QRegExp &exp = hilite.regexp;
		if (exp.pattern().isEmpty()) continue;
		int index = exp.indexIn(text);
		while (index >= 0) {
			int len = exp.matchedLength();
			if (len <= 0) break;
			setFormat(index, len, hilite.format);
			index = exp.indexIn(text, index + len);
		}
	}
}
